// HC11/CS8900 ethernet IP stack
//
// $Id: ether.c,v 1.7 2001/06/21 00:31:13 tom Exp tom $
//
// Copyright (C) 2000-1 Tom Parker (tom-hc11@wiresncode.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Author's note: This is free software, for commerical use I would
// appreciate some recognition though.  If you are using this and like
// it send me some email or a postcard.  Code changes for inclusion
// are also welcome.
//
// Tom Parker, PO Box 2165, Fitzroy, VIC, 3065, AUSTRALIA (tom-hc11@wiresncode.com)
//

#define M6811_CPU_CLOCK (16000000L)
#define M6811_DEF_BAUD  (0x30)

#define EXTERN

// m6811 includes
#include <sys/types.h>
#include <string.h>

#undef  __BYTE_ORDER
#define __BYTE_ORDER __BIG_ENDIAN

#define ntohs(tom) (tom)
#define htons(tom) (tom)

// linux includes
#include "linux.h"

// local includes
#include "m68hc11/sio.h"
#include "ports_def_f1.h"

#include "ntp.h"
#include "bootp.h"
#include "cs8900.h"

//-----------------------------------------------------------------
// Constants

#define BUFFERLEN      0x200
#define MACTEMPLATE    "Tom(0)"

#ifdef BUFFALO
#define PVBASE 0x7e00
#define PVRTI  0x27
#endif

#define ARPINTERVAL    5
#define MAXARPATTEMPTS 5

//-----------------------------------------------------------------
// Globals
static u_int8_t arpBuffer[sizeof(struct ether_header) + sizeof(struct ether_arp)];
static u_int8_t txBuffer[BUFFERLEN];
static u_int8_t ourMAC[ETHER_ADDR_LEN];
static u_int8_t ourIP[sizeof(struct in_addr)];
static u_int8_t ourBCAST[sizeof(struct in_addr)];

static u_int8_t arpMAC[ETHER_ADDR_LEN];
static u_int8_t arpIP[sizeof(struct in_addr)];
static u_int8_t arpTimeout;
static u_int8_t arpValid;

static u_int8_t rxBuffer[BUFFERLEN];

static u_int32_t currentTime;

static u_int8_t bootpPending;
static u_int8_t bootpTimeout;
static uint16_t timerTicks;

//-----------------------------------------------------------------
// Prototypes
void cs8900_send(u_int8_t* aBuffer, u_int16_t aLength);
void time_of_day(void) __attribute__((interrupt));
void cs8900_poll();

//-----------------------------------------------------------------
// serial library routines
//
void serial_send_debug(u_int8_t c)
{
#ifdef DEBUG
   serial_send(c);
#endif
}

void
serial_send_crlf()
{
   serial_send('\r');
   serial_send('\n');
}

void 
serial_send_string(char* aString)
{
   while(*aString != '\0')
   {
      serial_send(*aString);
      aString++;
   }
}

void
serial_send_int(u_int16_t aNumber)
{
   u_int16_t divisor = 10000;
   u_int16_t modulo = 0;
   u_int8_t firstDigit = 1;

   if (aNumber == 0)
   {
      serial_send('0');
   }
   else while (divisor > 0)
   {
      modulo = aNumber - (aNumber % divisor);
      if (modulo > 0 || !firstDigit)
      {
	 firstDigit = 0;
	 serial_send('0' + modulo / divisor);
	 aNumber -= modulo;
      }
      
      divisor /= 10;
   }
}

void
serial_send_hex_int(u_int16_t aNumber, u_int8_t aLeadingZeros)
{
   u_int16_t  divisor = 0x1000;
   u_int8_t   value = 0;
   u_int16_t  modulo = 0;
   u_int8_t   firstChar = 1;

   if (aNumber == 0)
   {
      if (aLeadingZeros) serial_send_string("0000"); else serial_send('0');
   }
   else while (divisor > 0)
   {
      modulo = aNumber - (aNumber % divisor);
      value =  modulo / divisor;
      if (value != 0 || aLeadingZeros || !firstChar)
      {
	 if (value != 0) firstChar = 0;
	 if (value < 10)
	 {
	    serial_send('0' + value);
	 }
	 else 
	 {
	    serial_send('a' - 10 + value);
	 }
      }
      aNumber -= modulo;
      divisor /= 0x10;
   }
}

void
serial_send_hex_byte(u_int8_t aNumber, u_int8_t aLeadingZeros)
{
   u_int8_t divisor = 0x10;
   u_int8_t value = 0;
   u_int8_t modulo = 0;
   u_int8_t firstChar = 1;

   if (aNumber == 0)
   {
      if (aLeadingZeros) serial_send_string("00"); else serial_send('0');
   }
   else while (divisor > 0)
   {
      modulo = aNumber - (aNumber % divisor);
      value =  modulo / divisor;
      if (value != 0 || aLeadingZeros || !firstChar)
      {
	 if (value != 0) firstChar = 0;
	 if (value < 10)
	 {
	    serial_send('0' + value);
	 }
	 else 
	 {
	    serial_send('a' - 10 + value);
	 }
      }
      aNumber -= modulo;
      divisor /= 0x10;
   }
}
//-----------------------------------------------------------------
// Time of day ISR and maintenance
// 
// ISR will fire every 2.048 ms.  time_of_day is hooked to the RTI
// Real Time Interrupt in vectors.s for standalone or in hc11_init
// for buffalo rom
//
void
time_of_day()
{
   // fire every second
   if (++timerTicks == 1000000/16384)
   {
      timerTicks = 0;
      currentTime++;
      serial_send_debug('T');

      // 1s timeout for bootp
      if (bootpPending) bootpTimeout = 1;

      // decrement timeout for ARP
      if (arpTimeout) arpTimeout--;
   }

   _io_ports[M6811_TFLG2] |= M6811_RTIF;
}

//-----------------------------------------------------------------
// dump rxBuffer
void
rxDump()
{
   const u_int8_t* rxPtr = rxBuffer;
   u_int16_t i = 0;

   serial_send_crlf();
	       
   for (i = 1; i <= BUFFERLEN; i++, rxPtr++)
   {
      serial_send_hex_byte(*rxPtr, 1);
      serial_send(' ');
      if (i % 16 == 0) serial_send_crlf();
   }

   serial_send_crlf();
}

//-----------------------------------------------------------------
// dump txBuffer
void
txDump()
{
   const u_int8_t* txPtr = txBuffer;
   u_int16_t i = 0;

   serial_send_crlf();
	       
   for (i = 1; i <= BUFFERLEN; i++, txPtr++)
   {
      serial_send_hex_byte(*txPtr, 1);
      serial_send(' ');
      if (i % 16 == 0) serial_send_crlf();
   }

   serial_send_crlf();
}

//-----------------------------------------------------------------
// ARP packet generation
void
send_arp(const unsigned short int aOperation, const u_int8_t* aDestMAC, const u_int8_t* aDestIP)
{
   struct ether_header* ethHdr = (struct ether_header*) arpBuffer;
   struct ether_arp* arpPkt = (struct ether_arp*) (arpBuffer + sizeof(struct ether_header));

   // can't send ARP if our address is not set
   if (bootpPending == 0)
   {
      serial_send_debug('a');

      // build MAC header
      memcpy((void *)ethHdr->ether_dhost, (void *)aDestMAC, ETHER_ADDR_LEN);
      ethHdr->ether_type = ETHERTYPE_ARP;
   
      // build ARP packet header
      arpPkt->ea_hdr.ar_hrd = ARPHRD_ETHER;
      arpPkt->ea_hdr.ar_pro = ETHERTYPE_IP;
      arpPkt->ea_hdr.ar_hln = ETHER_ADDR_LEN;
      arpPkt->ea_hdr.ar_pln = sizeof(struct in_addr);
      arpPkt->ea_hdr.ar_op = aOperation;

      switch(aOperation)
      {
         case ARPOP_REPLY :
         {
            serial_send_debug('1');

            // build arp reply
            memcpy(arpPkt->arp_sha, ourMAC, ETHER_ADDR_LEN);
            memcpy(arpPkt->arp_tha, aDestMAC, ETHER_ADDR_LEN);
            memcpy(arpPkt->arp_spa, ourIP, sizeof(struct in_addr));
            memcpy(arpPkt->arp_tpa, aDestIP, sizeof(struct in_addr));
            break;
         }
         case ARPOP_REQUEST :
         {
            serial_send_debug('2');

            // build arp request
            memcpy(arpPkt->arp_sha, ourMAC, ETHER_ADDR_LEN);
            memcpy(arpPkt->arp_tha, aDestMAC, ETHER_ADDR_LEN);
            memcpy(arpPkt->arp_spa, ourIP, sizeof(struct in_addr));
            memcpy(arpPkt->arp_tpa, aDestIP, sizeof(struct in_addr));

            // update table
            arpValid = 0;
            memcpy(arpMAC, aDestMAC, ETHER_ADDR_LEN);
            memcpy(arpIP, aDestIP, sizeof(struct in_addr));

            break;
         }
      }
    
      // send request/reply
      cs8900_send(arpBuffer, sizeof(struct ether_header) + sizeof(struct ether_arp));
   }
}

//-----------------------------------------------------------------
// send an ip packet (protocol, length & destination address are
// assumed to be filled out).  Packet is MAC size packet ready to go.
void
send_ip()
{
   struct ether_header* ethHdr = (struct ether_header*) txBuffer;
   struct ip* ipPkt = (struct ip*) (txBuffer + sizeof(struct ether_header));
   const u_int16_t* mem = (u_int16_t*) ipPkt;
   u_int32_t newcksum = 0;
   register int i = 0;

   serial_send_debug('i');

   ipPkt->ip_v = IPVERSION;
   ipPkt->ip_hl = 5;
   ipPkt->ip_tos = 0;
   ipPkt->ip_off = 0;

   // check bootp status
   if (bootpPending == 0)
   {
      u_int8_t arpRetryCount = 0;

      // address is assigned so use our IP & a specific MAC address
      memcpy(&ipPkt->ip_src, ourIP, sizeof(struct in_addr));

      if (memcmp(&ipPkt->ip_dst, arpIP, sizeof(struct in_addr)))
      {
         arpValid = 0;
      }

      arpTimeout = 0;
      while(arpValid != 1)
      {
         if (arpTimeout == 0)
         {
            const char destMAC[ETHER_ADDR_LEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
            
            if (arpRetryCount++ > MAXARPATTEMPTS)
            {
               serial_send_string("Aborting TX - ARP failed\r\n");
               
               return;
            }

            send_arp(ARPOP_REQUEST, destMAC, (u_int8_t *) &ipPkt->ip_dst);
            arpTimeout = ARPINTERVAL;
         }

         cs8900_poll();
      }
      memcpy(ethHdr->ether_dhost, arpMAC, ETHER_ADDR_LEN);
   }
   else
   {
      // use null IP address & bcast MAC address if bootp is not finished
      u_int8_t bcast[ETHER_ADDR_LEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

      memcpy((void *) ethHdr->ether_dhost, (void *) bcast, ETHER_ADDR_LEN);
      bzero((void *) &ipPkt->ip_src, sizeof(struct in_addr)); 
   }

   // checksum
   ipPkt->ip_sum = 0;
   for(i = 0; i < 10; i++, mem++)
   {
      newcksum += *mem;
   }
   ipPkt->ip_sum = htons((u_int16_t) ~((newcksum >> 16) + (newcksum & 0xffff)));

   // MAC
   ethHdr->ether_type = ETHERTYPE_IP;

   cs8900_send(txBuffer, ntohs(ipPkt->ip_len) + sizeof(struct ether_header));
}

//-----------------------------------------------------------------
// send a udp packet given the address, length & payload in txBuffer
void
send_udp(const u_int16_t aSourcePort, 
	 const u_int16_t aDestPort, 
	 const u_int8_t* aAddress, 
	 const u_int16_t aLength)
{
   struct ip* ipPkt = (struct ip*) (txBuffer + sizeof(struct ether_header));
   struct udphdr* udpPkt = (struct udphdr*) (txBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   const u_int16_t* mem = (u_int16_t*) &(ipPkt->ip_src);
   u_int32_t newcksum = IPPROTO_UDP + sizeof(struct udphdr) + aLength;
   register u_int16_t i = 0;
   u_int16_t length = aLength;

   serial_send_debug('u');

   // UDP header
   udpPkt->source = htons(aSourcePort);
   udpPkt->dest = htons(aDestPort);
   udpPkt->len = htons(length + sizeof(struct udphdr));
   udpPkt->check = 0;

   // destination IP
   memcpy(&ipPkt->ip_dst.s_addr, aAddress, sizeof(struct in_addr));

   // fill source address for checksum - use a zero ip address if we
   // do not have one
   if (bootpPending == 0)
   {
      memcpy(&ipPkt->ip_src, ourIP, sizeof(struct in_addr));
   }
   else
   {
      bzero((void *) &ipPkt->ip_src, sizeof(struct in_addr)); 
   }
   
   // packet length, protocol & id are the same
   ipPkt->ip_len = htons(length + sizeof(struct ip) + sizeof(struct udphdr));
   ipPkt->ip_p = IPPROTO_UDP;
   ipPkt->ip_id = 0;
   ipPkt->ip_ttl = 64;

   // checksum includes protocol (see above), udp length (see
   // above) and IP source & destination
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem);

   // packet checksum
   mem = (u_int16_t*) udpPkt;
   
   // pad odd packets
   if (length % 2)
   {
      *((u_int8_t *) mem + length) = 0;
   }

   // calc checksum (convert length to words & round odd bytes up)
   length = (length + 1) >> 1;
   for (i = 0; i < length; i++, mem++)
   {
      newcksum += ntohs(*mem);
   }

   // insert checksum & send
   udpPkt->check = htons((u_int16_t) ~((newcksum >> 16) + (newcksum & 0xffff)));
   send_ip();
}

//-----------------------------------------------------------------
// send a bootp packet given the address, length & payload in txBuffer
void
send_bootp()
{
   struct bootp* bootpPkt = (struct bootp*) (txBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));
   u_int8_t destAddr[sizeof(struct in_addr)] = {0xff, 0xff, 0xff, 0xff};
   const u_int8_t rfc1048cookie[sizeof(struct in_addr)] =  VM_RFC1048;

   serial_send_debug('b');

   // fill packet inc. our hardware address
   bzero((void *) bootpPkt, sizeof(struct bootp));
   bootpPkt->bp_op = BOOTREQUEST;
   bootpPkt->bp_htype = HTYPE_ETHERNET;
   bootpPkt->bp_hlen = ETHER_ADDR_LEN;
   bootpPkt->bp_flags = htons(0x8000); // broadcast
   memcpy(bootpPkt->bp_vend, rfc1048cookie, sizeof(struct in_addr));
   memcpy((void *) bootpPkt->bp_chaddr, (void *) ourMAC, ETHER_ADDR_LEN);

   bootpTimeout = 0;
   serial_send('.');
   
   send_udp(IPPORT_BOOTPC, IPPORT_BOOTPS, destAddr, sizeof(struct bootp));
}

//-----------------------------------------------------------------
// send an ICMP echo response using the received request data
void
send_icmp_echoreply()
{
   const struct ip* rxIpPkt = (struct ip*) (rxBuffer + sizeof(struct ether_header));
   const struct icmphdr* rxIcmpPkt = (struct icmphdr*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   struct ip* ipPkt = (struct ip*) (txBuffer + sizeof(struct ether_header));
   struct icmphdr* icmpPkt = (struct icmphdr*) (txBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   const u_int16_t* mem = (u_int16_t*) icmpPkt;
   u_int16_t icmpPktLen = ntohs(rxIpPkt->ip_len) - sizeof(struct ip);
   u_int32_t newcksum = 0;
   register u_int16_t i = 0;

   serial_send_debug('p');

   // create reply
   icmpPkt->type = ICMP_ECHOREPLY;
   icmpPkt->code = 0;
   icmpPkt->un.echo.id = rxIcmpPkt->un.echo.id;
   icmpPkt->un.echo.sequence = rxIcmpPkt->un.echo.sequence;

   // copy payload
   memcpy((u_int8_t*) icmpPkt + sizeof(struct icmphdr),
	  (u_int8_t*) rxIcmpPkt + sizeof(struct icmphdr),
	  icmpPktLen - sizeof(struct icmphdr));

   // checksum
   icmpPkt->checksum = 0;

   // pad odd packets
   if (icmpPktLen % 2)
   {
      *((u_int8_t *) mem + icmpPktLen) = 0;
   }

   // checksum (convert length to words & round odd bytes up)
   icmpPktLen = (icmpPktLen + 1) >> 1;
   for(i = 0; i < icmpPktLen; i++, mem++)
   {
      newcksum += ntohs(*mem);
   }
   icmpPkt->checksum = htons((u_int16_t) ~((newcksum >> 16) + (newcksum & 0xffff)));
   
   // replace destination IP with source of rx packet
   memcpy(&ipPkt->ip_dst, &rxIpPkt->ip_src, sizeof(struct in_addr));
   
   // packet length, protocol & id are the same
   ipPkt->ip_len = rxIpPkt->ip_len;
   ipPkt->ip_p = IPPROTO_ICMP;
   ipPkt->ip_id = rxIpPkt->ip_id;
   ipPkt->ip_ttl = rxIpPkt->ip_ttl;

   send_ip();
}

//-----------------------------------------------------------------
// ARP packet receive & action
void
receive_arp()
{
   const struct ether_arp* arpPkt = (struct ether_arp*) (rxBuffer + sizeof(struct ether_header));

   serial_send_debug('A');

   if (arpPkt->ea_hdr.ar_hrd == ARPHRD_ETHER && arpPkt->ea_hdr.ar_pro == ETHERTYPE_IP)
   {

      serial_send_debug('1');
      switch(arpPkt->ea_hdr.ar_op)
      {
	 case ARPOP_REQUEST:
	 {
#ifdef ARPDEBUG
	    serial_send_crlf();
	    serial_send_string("Got ARP request for ");
	    serial_send_int(arpPkt->arp_tpa[0]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_tpa[1]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_tpa[2]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_tpa[3]);
	    serial_send_crlf();
#endif
	    serial_send_debug('2');

	    // respond to requests for our ip address
	    if (!memcmp(ourIP, arpPkt->arp_tpa, sizeof(struct in_addr)))
	    {
	       serial_send_debug('3');
	       send_arp(ARPOP_REPLY, arpPkt->arp_sha, arpPkt->arp_spa);
	    }
	    break;
	 }
	 case ARPOP_REPLY:
	 {
#ifdef ARPDEBUG
            int i = 0;

	    serial_send_crlf();
	    serial_send_string("Got ARP reply for ");
	    serial_send_int(arpPkt->arp_spa[0]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_spa[1]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_spa[2]);
	    serial_send('.');
	    serial_send_int(arpPkt->arp_spa[3]);
	    serial_send_string(" MAC address is ");
            for(i = 0; i < 5; i++)
            {
               serial_send_hex_byte((u_int16_t) arpPkt->arp_sha[i], 0);
               serial_send(':');
            }
            serial_send_hex_byte((u_int16_t) arpPkt->arp_sha[5], 0);
	    serial_send_crlf();

	    serial_send_int(arpIP[0]);
	    serial_send('.');
	    serial_send_int(arpIP[1]);
	    serial_send('.');
	    serial_send_int(arpIP[2]);
	    serial_send('.');
	    serial_send_int(arpIP[3]);
	    serial_send_crlf();
#endif
	    serial_send_debug('3');

            // check this the correct response
            if (!memcmp(arpPkt->arp_spa, arpIP, sizeof(struct in_addr)))
            {
               // update table
               memcpy(arpMAC, arpPkt->arp_sha, ETHER_ADDR_LEN);
               arpValid = 1;
               serial_send_debug('4');
            }
	    break;
	 }
      }
   }
}
//-----------------------------------------------------------------
// ICMP packet receive & action
void
receive_icmp(const uint16_t aLength)
{
   const struct icmphdr* icmpPkt = (struct icmphdr*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   const u_int16_t* mem = (u_int16_t*) icmpPkt;
   u_int32_t newcksum = 0;
   register u_int16_t i = 0;
   u_int16_t length = aLength;

   serial_send_debug('P');

   // pad odd packets
   if (length % 2)
   {
      *((u_int8_t *) mem + length) = 0;
   }

   // checksum (convert length to words & round odd bytes up)
   length = (length + 1) >> 1;
   for (i = 0; i < length; i++, mem++)
   {
      // skip existing checksum
      if (i != 1)
      {
	 newcksum += ntohs(*mem);
      }
   }

   newcksum = ~((newcksum >> 16) + (newcksum & 0xffff));
   if ((newcksum & 0xffff) == ntohs(icmpPkt->checksum))
   {
      switch(icmpPkt->type)
      {
	 case ICMP_ECHO:
	    serial_send_debug('1');
	    send_icmp_echoreply();
	    break;
      }
   }
   else
   {
      serial_send_string("ICMP checksum error 0x");
      serial_send_hex_int(newcksum,1);
      serial_send_string(" != 0x");
      serial_send_hex_int(ntohs(icmpPkt->checksum), 1);
      serial_send_crlf();
      rxDump();
   }
}

//-----------------------------------------------------------------
// UDP packet receive & action
void
receive_ntp(const u_int16_t aLength)
{
   serial_send_debug('N');

   if (aLength >= sizeof(struct ntp))
   {
      struct ntp* ntpPkt = (struct ntp*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));

      currentTime = ntpPkt->xmt.l_ui;
   }
   else
   {
      serial_send_string("\r\nUndersize NTP packet received ");
      serial_send_int(aLength);
      serial_send_string(" < ");
      serial_send_int(sizeof(struct ntp));
      serial_send_crlf();
   }
}

//-----------------------------------------------------------------
// UDP packet receive & action
void
receive_bootp(const u_int16_t aLength)
{
   struct bootp* bootpPkt = (struct bootp*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));

   serial_send_debug('B');

   // only interested in responses
   if (bootpPkt->bp_op == BOOTREPLY && 
       bootpPkt->bp_htype == HTYPE_ETHERNET &&
       bootpPending == 1 &&
       !memcmp((void *) bootpPkt->bp_chaddr, (void *) ourMAC, ETHER_ADDR_LEN))
   {
      const u_int8_t rfc1048cookie[sizeof(struct in_addr)] =  VM_RFC1048;

      // get our IP address
      memcpy(ourIP, &bootpPkt->bp_yiaddr.s_addr, sizeof(struct in_addr));

      serial_send_string(" got response - addr=");
      serial_send_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[0]);
      serial_send('.');
      serial_send_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[1]);
      serial_send('.');
      serial_send_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[2]);
      serial_send('.');
      serial_send_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[3]);
      
      // check for options
      if (!memcmp(bootpPkt->bp_vend, rfc1048cookie, sizeof(struct in_addr)))
      {
	 u_int8_t* mem = bootpPkt->bp_vend + sizeof(struct in_addr);

	 while(*mem !=  TAG_END)
	 {
	    switch(*mem)
	    {
	       case TAG_IP_BROADCAST:
		  if (*(mem + 1) == 4)
		  {
		     // get broadcast address
		     memcpy(ourBCAST, mem + 2, sizeof(struct in_addr));

		     serial_send_string(", broadcast=");
		     serial_send_int(*(mem + 2));
		     serial_send('.');
		     serial_send_int(*(mem + 3));
		     serial_send('.');
		     serial_send_int(*(mem + 4));
		     serial_send('.');
		     serial_send_int(*(mem + 5));
		  }
		  break;
	       default:
		  // ignore for now
	    }

	    // skip TAG + length fields + option length
	    mem += *(mem + 1) + 2;
	 }
      }
      serial_send_crlf();

      // release flag
      bootpPending--;
   }
}

//-----------------------------------------------------------------
// UDP packet receive & action
void
receive_udp(const u_int16_t aLength)
{
   struct udphdr* udpPkt = (struct udphdr*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   u_int16_t length = aLength;

   serial_send_debug('U');

   if (length == ntohs(udpPkt->len))
   {
      const struct ip* ipPkt = (struct ip*) (rxBuffer + sizeof(struct ether_header));
      const u_int16_t* mem = (u_int16_t*) &(ipPkt->ip_src);
      u_int32_t newcksum = IPPROTO_UDP + length;
      register u_int16_t i = 0;

      // checksum includes protocol (see above), udp length (see
      // above) and IP source & destination
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem);

      // packet checksum
      mem = (u_int16_t*) udpPkt;

      // pad odd packets
      if (length % 2)
      {
	 *((u_int8_t *) mem + length) = 0;
      }

      // calc checksum (convert length to words & round odd bytes up)
      length = (length + 1) >> 1;
      for (i = 0; i < length; i++, mem++)
      {
	 // skip existing checksum
	 if (i != 3)
	 {
	    newcksum += ntohs(*mem);
	 }
      }
   
      newcksum = ~((newcksum >> 16) + (newcksum & 0xffff));
      if ((newcksum & 0xffff) == ntohs(udpPkt->check))
      {
	 switch(ntohs(udpPkt->dest))
	 {
	    case 123: // NTP
	       serial_send_debug('1');
	       receive_ntp(ntohs(udpPkt->len));
	       break;
	    case IPPORT_BOOTPC:
	       serial_send_debug('2');
	       receive_bootp(ntohs(udpPkt->len));
	       break;
	 }
      }
      else
      {
	 serial_send_string("\r\nUDP checksum error 0x");
	 serial_send_hex_int(newcksum,1);
	 serial_send_string(" != 0x");
	 serial_send_hex_int(ntohs(udpPkt->check), 1);
	 serial_send_crlf();
	 rxDump();
      }
   }
   else
   {
      serial_send_string("\r\nUDP length mismatch ");
      serial_send_int(length);
      serial_send_string(" != ");
      serial_send_int(udpPkt->len);
      serial_send_crlf();
   }
}

//-----------------------------------------------------------------
// IP packet receive & action
void
receive_ip()
{
   const struct ip* ipPkt = (struct ip*) (rxBuffer + sizeof(struct ether_header));

   serial_send_debug('I');

   // extract IPv4, non fragmented, optionless packets with a valid length
   // destined for us (if bootp is done) or broadcast
   if ((ipPkt->ip_v == IPVERSION && 
	(ipPkt->ip_off & IP_OFFMASK) == 0 && 
	ntohs(ipPkt->ip_len) > 20 &&
	ipPkt->ip_hl == 5) &&
       ((!memcmp(ourIP, &ipPkt->ip_dst, sizeof(struct in_addr)) && bootpPending == 0) ||
	(!memcmp(ourBCAST, &ipPkt->ip_dst, sizeof(struct in_addr)))))
   {
      const struct ether_header* ethHdr = (struct ether_header*) rxBuffer;
      u_int32_t newcksum = 0;
      u_int16_t* mem = (u_int16_t*) ipPkt;
      register u_int16_t i = 0;

      // save MAC address and IP address to prevent ARP
      memcpy(arpMAC, ethHdr->ether_shost, ETHER_ADDR_LEN);
      memcpy(arpIP, &ipPkt->ip_src, sizeof(struct in_addr));
      arpValid = 1;

      serial_send_debug('1');
      
      // checksum
      for (i = 0; i < 10; i++, mem++)
      {
	 // skip existing checksum
	 if (i != 5)
	 {
	    newcksum += *mem;
	 }
      }
      newcksum = ~((newcksum >> 16) + (newcksum & 0xffff));
      
      if ((newcksum & 0xffff) == ipPkt->ip_sum)
      {
	 serial_send_debug('2');

	 // Valid packet
	 switch(ipPkt->ip_p)
	 {
	    case IPPROTO_ICMP:
	    {
	       receive_icmp(ntohs(ipPkt->ip_len) - 20);
	       break;
	    }
	    case IPPROTO_UDP:
	    {
	       receive_udp(ntohs(ipPkt->ip_len) - 20);
	       break;
	    }
	 }
      }
      else
      {
	 serial_send_string("IP checksum error 0x");
	 serial_send_hex_int(newcksum,1);
	 serial_send_string(" != 0x");
	 serial_send_hex_int(ipPkt->ip_sum,1);
	 serial_send_crlf();
      }
   }
   
}

//-----------------------------------------------------------------
// cs8900 library routines
//
static inline u_int16_t
cs8900_pp_reg_read(u_int16_t reg)
{
   _io_ports[CS8900_PP_ADDR_LSB] = (u_int8_t) reg & 0xff;
   _io_ports[CS8900_PP_ADDR_MSB] = (u_int8_t) (reg >> 8) & 0xff;

   return (_io_ports[CS8900_PPDATA_PORT0_LSB] | (_io_ports[CS8900_PPDATA_PORT0_MSB] << 8));
}

void
cs8900_pp_reg_write(u_int16_t reg, u_int16_t data)
{
   _io_ports[CS8900_PP_ADDR_LSB] = (u_int8_t) reg & 0xff;
   _io_ports[CS8900_PP_ADDR_MSB] = (u_int8_t) (reg >> 8) & 0xff;

   _io_ports[CS8900_PPDATA_PORT0_LSB] = data & 0xff;
   _io_ports[CS8900_PPDATA_PORT0_MSB] = (data >> 8) & 0xff;
}

void
cs8900_init(u_int8_t aNodeId)
{
   u_int8_t i = 0;

   // initial IP broadcast address
   memcpy(ourBCAST, "\xff\xff\xff\xff", sizeof(struct in_addr));

   // MAC address
   memcpy(ourMAC, MACTEMPLATE, ETHER_ADDR_LEN);
   if (aNodeId > 9)
   {
      ourMAC[4] = aNodeId + 'a' - 10;
   }
   else
   {
      ourMAC[4] = aNodeId + '0';
   }

   serial_send_string("             - set MAC Address to ");
   for(i = 0; i < 5; i++)
   {
      serial_send_hex_byte((u_int16_t) ourMAC[i], 0);
      serial_send(':');
   }
   serial_send_hex_byte((u_int16_t) ourMAC[5], 0);
   serial_send_crlf();

   while(!cs8900_pp_reg_read(PP_SelfST) & INIT_DONE)
   {
      serial_send('.');
   }

   // enable full duplex
   cs8900_pp_reg_write(PP_TestCTL, TESTCTL_FDX);

   // accept addressed and broadcast valid frames 
   cs8900_pp_reg_write(PP_RxCTL, RX_OK_ACCEPT     |
		       RX_IA_ACCEPT               |
		       RX_BROADCAST_ACCEPT);

   // set our MAC address
   cs8900_pp_reg_write(PP_IA,   ourMAC[1] << 8 | ourMAC[0]);
   cs8900_pp_reg_write(PP_IA+2, ourMAC[3] << 8 | ourMAC[2]);
   cs8900_pp_reg_write(PP_IA+4, ourMAC[5] << 8 | ourMAC[4]);

   // enable receiver and transmitter, force interface to 10baseT
   cs8900_pp_reg_write(PP_LineCTL, SERIAL_RX_ON | SERIAL_TX_ON);
}

int
cs8900_check()
{
   u_int16_t manufId = 0;
   u_int16_t prodId = 0;
   u_int8_t  retVal = 0;

   manufId = cs8900_pp_reg_read(PP_ChipID);
   if (manufId == 0x630e)
   {
      prodId = cs8900_pp_reg_read(PP_ChipID + 2);

      serial_send_string("Found CS8900 - manufacturer ID 0x");
      serial_send_hex_int(manufId, 1);
      serial_send_crlf();
      serial_send_string("             - product ID 0x");
      serial_send_hex_int(prodId, 1);
      serial_send_crlf();

      retVal = 1;
   }
   
   return retVal;
}

void
cs8900_send(u_int8_t* aBuffer, u_int16_t aLength)
{
   register u_int16_t i = 0;
   struct ether_header* ethHdr = (struct ether_header*) aBuffer;
   u_int8_t* txPtr = aBuffer;

   serial_send_debug('s');

   // insert our MAC address
   memcpy((void *) ethHdr->ether_shost, (void *) ourMAC, ETHER_ADDR_LEN);
   
   // start transmission only when all bytes are available
   _io_ports[CS8900_TxCMD_MSB] = 0;
   _io_ports[CS8900_TxCMD_LSB] = 0xc0;

#ifdef TXDEBUG
   serial_send_crlf();
   serial_send_string("Sending ");
   serial_send_int(aLength);
   serial_send_crlf();
#endif

   // write packet length
   _io_ports[CS8900_TxLength_MSB] = ((aLength >> 8) & 0xff);
   _io_ports[CS8900_TxLength_LSB] = (aLength & 0xff);

   // wait for packet space : TODO - TIMEOUT
   while (!(cs8900_pp_reg_read(PP_BusST) & READY_FOR_TX_NOW)) {};

   // pad odd packets
   if (aLength % 2)
   {
      *((u_int8_t *) aBuffer + aLength) = 0;
   }

   // write to xmit buffer (convert length to words & round odd bytes up)
   aLength = (aLength + 1) >> 1;
   for (i = 1; i <= aLength; i++)
   {
      _io_ports[CS8900_DATA_PORT0_LSB] = *txPtr++;
      _io_ports[CS8900_DATA_PORT0_MSB] = *txPtr++;

#ifdef TXDEBUG
      serial_send_hex_byte(*txPtr, 1);
      serial_send(' ');
      if (i % 16 == 0) serial_send_crlf();
#endif
   }
   
#ifdef TXDEBUG
   serial_send_crlf();
#endif
}


void
cs8900_receive_packet()
{
   u_int16_t rxStatus = 0;
   u_int16_t rxLength = 0;
   register u_int16_t i = 0;
   u_int8_t store = 1;
   register u_int8_t thisbyte = 0;
   struct ether_header *ethHdr = (struct ether_header *) rxBuffer;

   serial_send_debug('R');

   // read length and status
   rxStatus = (_io_ports[CS8900_DATA_PORT0_MSB] << 8) | _io_ports[CS8900_DATA_PORT0_LSB];
   rxLength = (_io_ports[CS8900_DATA_PORT0_MSB] << 8) | _io_ports[CS8900_DATA_PORT0_LSB];

#ifdef RXDEBUG
   serial_send_crlf();
   serial_send_string("Receive length = ");
   serial_send_int(rxLength);
   serial_send_crlf();
#endif
   // discard long packets
   if (rxLength > BUFFERLEN) store = 0;

   // read packet in words
   for(i = 0; i < rxLength;i += 2)
   {
#ifdef RXDEBUG
      if (i % 16 == 0 && i != 0) serial_send_crlf();
#endif

      thisbyte = _io_ports[CS8900_DATA_PORT0_LSB];
      if (store)
      {
	 rxBuffer[i] = thisbyte;
      }

#ifdef RXDEBUG
      serial_send_hex_byte(thisbyte, 1);
      serial_send(' ');
#endif
      thisbyte = _io_ports[CS8900_DATA_PORT0_MSB];
      if (store)
      {
	 rxBuffer[i+1]   = thisbyte;
      }
#ifdef RXDEBUG
      serial_send_hex_byte(thisbyte, 1);
      serial_send(' ');
#endif
   }
#ifdef RXDEBUG
   serial_send_crlf();
#endif

   // analyse packet
   switch(ethHdr->ether_type)
   {
      case ETHERTYPE_ARP:
      {
	 serial_send_debug('1');
	 receive_arp();
	 break;
      }
      case ETHERTYPE_IP:
      {
	 serial_send_debug('2');
	 receive_ip();
	 break;
      }
   }
#ifdef RXDEBUG
   serial_send_crlf();
#endif
#ifdef DEBUG
   serial_send_crlf();
#endif
}

//-----------------------------------------------------------------
// poll the cs8900 for received packets
void 
cs8900_poll()
{
   u_int16_t rxEvent = cs8900_pp_reg_read(PP_RxEvent);
   
   if ( rxEvent & RX_OK)
   {
      cs8900_receive_packet();
   }
}


//-----------------------------------------------------------------
// HC11 initialisation
void
hc11_init()
{
#ifdef BUFFALO
   u_int8_t* memptr = (u_int8_t*) PVBASE + PVRTI;

   // fill buffalo psuedo vector table with a jump to the time_of_day handler
   *(memptr++) = 0x7e;
   *(memptr++) = (u_int16_t) time_of_day >> 8;
   *(memptr++) = (u_int16_t) time_of_day & 0xff;
#endif

   // CS follows E with 1 cycle stretch
   _io_ports[M6811F1_CSSTRH] |=  M6811F1_IO1SB;
   _io_ports[M6811F1_CSSTRH] &= ~M6811F1_IO1SA;
   
   // enable CS1, disable CS2
   _io_ports[M6811F1_CSCTL] |= M6811F1_IO1EN;
   _io_ports[M6811F1_CSCTL] &= ~M6811F1_IO2EN;
   
   // A/D converter off
   _io_ports[M6811_OPTION] |= ~M6811_ADPU;

   // Port G outputs
   _io_ports[M6811F1_DDRG] &= 0x0f;

   // config
   _io_ports[M6811_PACTL] |= M6811_RTR0;
   _io_ports[M6811_PACTL] |= M6811_RTR1;
   _io_ports[M6811_TMSK2] |= M6811_RTII;
}

int
main()
{
   u_int8_t nodeId = 0;
   bootpPending = 0;
   bootpTimeout = 0;
   timerTicks = 0;
   arpValid = 0;

   serial_init();
   hc11_init();

   // node id is set by lower 4 bits of port E
   nodeId = _io_ports[M6811_PORTE] & 0xf;

   serial_send_string("\r\nCS8900 IP (node = ");
   serial_send_hex_byte(nodeId, 0);
   serial_send_string(") $Revision: 1.7 $ - ");
   serial_send_string(__DATE__ " " __TIME__ "\r\n\r\n");

   if (!cs8900_check())
   {
      serial_send_string("Failed to find CS8900\n\r");

      while(1);
   }

   cs8900_init(nodeId);

#ifdef USEBOOTP
   // flag bootp as pending
   bootpPending = 1;
   bootpTimeout = 1;
   serial_send_string("\r\nBootp .");
#else
   // not using bootp so fix our IP address
   ourIP[0] = IPADDRESS_0;
   ourIP[1] = IPADDRESS_1;
   ourIP[2] = IPADDRESS_2;
   ourIP[3] = 200 + nodeId;
   serial_send_string("             - fixed ip address is ");
   serial_send_int(ourIP[0]);
   serial_send('.');
   serial_send_int(ourIP[1]);
   serial_send('.');
   serial_send_int(ourIP[2]);
   serial_send('.');
   serial_send_int(ourIP[3]);
   serial_send_crlf();

   // class C (a.b.c.255)
   ourBCAST[0] = IPADDRESS_0;
   ourBCAST[1] = IPADDRESS_1;
   ourBCAST[2] = IPADDRESS_2;
   ourBCAST[3] = 255;
   serial_send_string("             - broadcast address is ");
   serial_send_int(ourBCAST[0]);
   serial_send('.');
   serial_send_int(ourBCAST[1]);
   serial_send('.');
   serial_send_int(ourBCAST[2]);
   serial_send('.');
   serial_send_int(ourBCAST[3]);
   serial_send_crlf();
#endif

   // wait for rx event
   while (1)
   {
      if (bootpPending && bootpTimeout)
      {
	 // send bootp packet
	 send_bootp();
      }

// This is an example of a board initiated transmit that will generate ARP requests etc.
#if 0
      if (!bootpPending)
      {
         const u_int8_t destIP[sizeof(struct in_addr)] = {192, 168, 0, 99};
         
         send_udp(IPPORT_DISCARD, IPPORT_DISCARD, destIP, sizeof(struct bootp));
      }
#endif
      
      // check for received packet
      cs8900_poll();
   }

   return(0);
}


