/* m68hc11/sio.h -- Utility methods to read/write the SIO
   Copyright 1999, 2000 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GDB, GAS, and the GNU binutils.

GDB, GAS, and the GNU binutils are free software; you can redistribute
them and/or modify them under the terms of the GNU General Public
License as published by the Free Software Foundation; either version
1, or (at your option) any later version.

GDB, GAS, and the GNU binutils are distributed in the hope that they
will be useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#ifndef _M68HC11_SIO_H
#define _M68HC11_SIO_H

#include <m68hc11/param.h>
#include <m68hc11/ports.h>

#ifndef EXTERN
#define EXTERN extern
#endif

#ifdef MC6811

EXTERN inline void
serial_init (void)
{
  _io_ports[M6811_BAUD] = M6811_DEF_BAUD;

  /* Setup character format 1 start, 8-bits, 1 stop.  */
  _io_ports[M6811_SCCR1] = 0;

  /* Enable reciever and transmitter.  */
  _io_ports[M6811_SCCR2] = 0xc;
}

/* Return != 0 if there is something to read on the serial line.  */
EXTERN inline unsigned char
serial_receive_pending (void)
{
  return _io_ports[M6811_SCSR] & M6811_RDRF;
}

/* Wait until the SIO has finished to send the character.  */
EXTERN inline void
serial_flush (void)
{
  while (!(_io_ports[M6811_SCSR] & M6811_TDRE))
    cop_optional_reset ();
}

/* Send the character on the serial line.  */
EXTERN inline void
serial_send (char c)
{
  serial_flush ();
  _io_ports[M6811_SCDR] = c;
  _io_ports[M6811_SCCR2] |= M6811_TE;
}

/* Wait for a character on the serial line and return it.  */
EXTERN inline unsigned char
serial_recv (void)
{
  while (!(_io_ports[M6811_SCSR] & M6811_RDRF))
    cop_optional_reset ();

  return _io_ports[M6811_SCDR];
}

EXTERN void serial_print (const char *msg);
EXTERN void serial_getline (char *buf);

#endif /* MC6811 */
#endif /* _M68HC11_SIO_H */

